/*
 * include/asm-microblaze/uaccess.h
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2008 Michal Simek
 * Copyright (C) 2008 PetaLogix
 * Copyright (C) 2006 Atmark Techno, Inc.
 */

#ifndef _ASM_MICROBLAZE_UACCESS_H
#define _ASM_MICROBLAZE_UACCESS_H

#include <linux/errno.h>
#include <asm/segment.h>
#include <linux/autoconf.h>
#include <asm/string.h>

#define VERIFY_READ	0
#define VERIFY_WRITE	1

#ifndef CONFIG_MMU

/* Check against bounds of physical memory */
static inline int ___range_ok(unsigned long addr, unsigned long size)
{
	return ((addr < CONFIG_XILINX_ERAM_START) ||
		((addr + size) >= (CONFIG_XILINX_ERAM_START + 
				   CONFIG_XILINX_ERAM_SIZE)));
}

#define __range_ok(addr, size) ___range_ok((unsigned long)(addr), (unsigned long)(size))

#define access_ok(type, addr, size) (__range_ok((addr), (size)) == 0)
#define __access_ok(add, size) (__range_ok((addr), (size)) == 0)

extern inline int bad_user_access_length (void)
{
	return 0;
}

#define __get_user(var, ptr)						      \
  ({									      \
	  int __gu_err = 0;						      \
	  unsigned long __gu_val;					      \
	  unsigned long long __gu_val2;					      \
	  switch (sizeof (*(ptr))) {					      \
	  case 1:							      \
	  case 2:							      \
	  case 4:							      \
		  __gu_val = *(ptr);					      \
		  break;						      \
	  case 8:							      \
		  memcpy(&__gu_val2, ptr, sizeof(__gu_val2));		      \
		  break;						      \
	  default:							      \
		  __gu_val = 0;						      \
		  __gu_err = __get_user_bad ();				      \
		  break;						      \
	  }								      \
	  (var) = __gu_val;						      \
	  __gu_err;							      \
  })
#define __get_user_bad()	(bad_user_access_length (), (-EFAULT))

#define __put_user(var, ptr)						      \
  ({									      \
	  int __pu_err = 0;						      \
	  switch (sizeof (*(ptr))) {					      \
	  case 1:							      \
	  case 2:							      \
	  case 4:							      \
		  *(ptr) = (var);					      \
		  break;						      \
	  case 8: {							      \
	  	  typeof(*(ptr)) __pu_val = var;				      \
		  memcpy(ptr, &__pu_val, sizeof(__pu_val));		      \
		  }							      \
		  break;						      \
	  default:							      \
		  __pu_err = __put_user_bad ();				      \
		  break;						      \
	  }								      \
	  __pu_err;							      \
  })
#define __put_user_bad()	(bad_user_access_length (), (-EFAULT))

#define put_user(x, ptr)	__put_user(x, ptr)
#define get_user(x, ptr)	__get_user(x, ptr)

#define copy_to_user(to,from,n)			(memcpy(to, from, n), 0)
#define copy_from_user(to,from,n)		(memcpy(to, from, n), 0)

#define __copy_to_user(to,from,n)		(copy_to_user(to,from,n))
#define __copy_from_user(to,from,n)		(copy_from_user(to,from,n))
#define __copy_to_user_inatomic(to,from,n)	(__copy_to_user(to,from,n))
#define __copy_from_user_inatomic(to,from,n)	(__copy_from_user(to,from,n))

#define __clear_user(addr,n)		(memset((void *)addr, 0, n), 0)

static inline unsigned long clear_user(void *addr, unsigned long size)
{
 	if (access_ok(VERIFY_WRITE, addr, size))
 		size = __clear_user(addr, size);
 	return size;
}

/* Returns 0 if exception not found and fixup otherwise.  */
extern unsigned long search_exception_table(unsigned long);

extern long strncpy_from_user(char *dst, const char *src, long count);
extern long strnlen_user(const char *src, long count);


#else /* CONFIG_MMU */

#include <linux/mm.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/mmu.h>


/*
 * Address is valid if:
 *  - "addr", "addr + size" and "size" are all below the limit
 */

#define access_ok(type, addr, size) \
	(get_fs().seg > (((unsigned long) addr) | (size) | ((unsigned long)(addr) + (size))))

//	 || printk("access_ok failed for %s at 0x%08lx (size %d), seg 0x%08x\n", type?"WRITE":"READ",addr,size,get_fs().seg))

/*
 * All the __XXX versions macros/functions below do not perform
 * access checking. It is assumed that the necessary checks have been
 * already performed before the finction (macro) is called.
 */

#define get_user(x, ptr) \
({					\
  __typeof__(*(ptr)) *__ptr = ptr;	\
  access_ok(VERIFY_READ, __ptr, sizeof(*(__ptr)))?__get_user(x,__ptr): -EFAULT;	\
})

#define put_user(x, ptr) \
({					\
  __typeof__(*(ptr)) *__ptr = ptr;	\
  access_ok(VERIFY_WRITE, __ptr, sizeof(*(__ptr)))?__put_user(x,__ptr): -EFAULT;	\
})


#define __get_user(x,ptr)	\
({				\
  unsigned long __gu_val;	\
  unsigned long __gu_ptr = (unsigned long)(ptr);	\
  long __gu_err;					\
  switch(sizeof(*(ptr))) {				\
    case 1: __get_user_asm("lbu", __gu_ptr, __gu_val, __gu_err); break;	\
    case 2: __get_user_asm("lhu", __gu_ptr, __gu_val, __gu_err); break;	\
    case 4: __get_user_asm("lw", __gu_ptr, __gu_val, __gu_err); break;	\
    default: __gu_val = 0; __gu_err = -EINVAL;				\
  }									\
  x = (__typeof__(*(ptr))) __gu_val;					\
  __gu_err;			\
})

#define __get_user_asm(insn, __gu_ptr, __gu_val, __gu_err)	\
({							\
	__asm__ __volatile__ (				\
	"1:\t" insn "\t%1,%2,r0\n"			\
	"\taddk\t%0,r0,r0\n"				\
	"2:\n"						\
	"\t.section .fixup,\"ax\"\n"			\
	"3:\tbrid\t2b\n"				\
	"\taddik\t%0,r0,%3\n"				\
	"\t.previous\n"					\
	"\t.section __ex_table,\"a\"\n"			\
	"\t.word\t1b,3b\n"				\
	"\t.previous\n"					\
	: "=r"(__gu_err), "=r"(__gu_val)		\
	: "r"(__gu_ptr), "i"(-EFAULT)			\
	);						\
})

#define __put_user(x,ptr)	\
({				\
  __typeof__(*(ptr)) __gu_val = x;	\
  unsigned long __gu_ptr = (unsigned long)(ptr);	\
  long __gu_err;					\
  switch(sizeof(__gu_val)) {				\
    case 1: __put_user_asm("sb", __gu_ptr, __gu_val, __gu_err); break;	\
    case 2: __put_user_asm("sh", __gu_ptr, __gu_val, __gu_err); break;	\
    case 4: __put_user_asm("sw", __gu_ptr, __gu_val, __gu_err); break;	\
    case 8: __put_user_asm_8(__gu_ptr, __gu_val, __gu_err); break;	\
    default: __gu_err = -EINVAL;					\
  }				\
  __gu_err;			\
})

#define __put_user_asm_8(__gu_ptr, __gu_val, __gu_err)	\
({							\
	__asm__ __volatile__ (				\
	"\tlwi\t%0,%1,0\n"				\
	"1:\tswi\t%0,%2,0\n"				\
	"\tlwi\t%0,%1,4\n"				\
	"2:\tswi\t%0,%2,4\n"				\
	"\taddk\t%0,r0,r0\n"				\
	"3:\n"						\
	"\t.section .fixup,\"ax\"\n"			\
	"4:\tbrid\t3b\n"				\
	"\taddik\t%0,r0,%3\n"				\
	"\t.previous\n"					\
	"\t.section __ex_table,\"a\"\n"			\
	"\t.word\t1b,4b,2b,4b\n"			\
	"\t.previous\n"					\
	: "=&r"(__gu_err)				\
	: "r"(&__gu_val), "r"(__gu_ptr), "i"(-EFAULT)	\
	);						\
})

#define __put_user_asm(insn, __gu_ptr, __gu_val, __gu_err)	\
({							\
	__asm__ __volatile__ (				\
	"1:\t" insn "\t%1,%2,r0\n"			\
	"\taddk\t%0,r0,r0\n"				\
	"2:\n"						\
	"\t.section .fixup,\"ax\"\n"			\
	"3:\tbrid\t2b\n"				\
	"\taddik\t%0,r0,%3\n"				\
	"\t.previous\n"					\
	"\t.section __ex_table,\"a\"\n"			\
	"\t.word\t1b,3b\n"				\
	"\t.previous\n"					\
	: "=r"(__gu_err)				\
	: "r"(__gu_val), "r"(__gu_ptr), "i"(-EFAULT)	\
	);						\
})


extern int __copy_user(char *to, char *from, int count);

/*
 * Return: number of not copied bytes, i.e. 0 if OK or non-zero if fail.
 */

static inline int clear_user(char *to, int size)
{
	if (size && access_ok(VERIFY_WRITE,to,size)) {
		__asm__ __volatile__ (
		"1:\n"
		"\tsb\tr0,%2,r0\n"
		"\taddik\t%0,%0,-1\n"
		"\tbneid\t%0,1b\n"
		"\taddik\t%2,%2,1\n"
		"2:\n"
		"\t.section __ex_table,\"a\"\n"
		"\t.word\t1b,2b\n"
		"\t.section .text\n"
		: "=r"(size)
		: "0"(size), "r"(to)
		);
	}
	return size;
}


#define copy_to_user(to, from, n) (access_ok(VERIFY_WRITE, to, n)? __copy_user((char *)to, (char *)from, n): -EFAULT)
#define copy_from_user(to, from, n) (access_ok(VERIFY_READ, from, n)? __copy_user((char *)to, (char *)from, n): -EFAULT)
#define __copy_to_user(to,from,n) copy_to_user(to,from,n)
#define __copy_from_user(to,from,n) copy_from_user(to,from,n)
#define __copy_to_user_inatomic(to,from,n) copy_to_user(to,from,n)
#define __copy_from_user_inatomic(to,from,n) copy_from_user(to,from,n)


extern int __strncpy_user(char *to, const char __user *from, int len);
extern int __strnlen_user(const char __user *sstr, int len);

#define strncpy_from_user(to,from,len)	(access_ok(VERIFY_READ,from,1)?__strncpy_user(to,from,len):-EFAULT)
#define strnlen_user(str,len)	(access_ok(VERIFY_READ,str,1)?__strnlen_user(str,len):0)

#endif /* CONFIG_MMU */

/*
 * The exception table consists of pairs of addresses: the first is the
 * address of an instruction that is allowed to fail, and the second is
 * the address at which the program should continue.  No registers are
 * modified, so it is entirely up to the continuation code to figure out
 * what to do.
 *
 * All the routines below use bits of fixup code that are out of line
 * with the main instruction path.  This means when everything is well,
 * we don't even have to jump over them.  Further, they do not intrude
 * on our cache or tlb entries.
 */
struct exception_table_entry
{
	unsigned long insn, fixup;
};

#endif /* _ASM_MICROBLAZE_UACCESS_H */
